from collections.abc import Callable, Generator
from typing import Any, Self

import bson

from .. import compat

if compat.IS_PYDANTIC_V2:
    from pydantic_core import CoreSchema, core_schema


class PyObjectId(bson.ObjectId):
    if compat.IS_PYDANTIC_V2:

        @classmethod
        def validate(
            cls, v: bson.ObjectId | bytes, _: core_schema.ValidationInfo = None
        ) -> bson.ObjectId:
            return cls._validate(v)

        @classmethod
        def __get_pydantic_core_schema__(cls, source_type: type[Self]) -> CoreSchema:
            return core_schema.with_info_plain_validator_function(
                cls.validate, serialization=core_schema.to_string_ser_schema()
            )

        @classmethod
        def __get_pydantic_json_schema__(
            cls,
            schema: core_schema.CoreSchema,
            handler: Callable[[Any], core_schema.CoreSchema],
        ) -> CoreSchema:
            return dict(type="string")

    else:

        @classmethod
        def __get_validators__(cls) -> Generator[Callable[[Any], bson.ObjectId], None, None]:
            yield cls.validate

        @classmethod
        def validate(cls, v: bson.ObjectId | bytes) -> bson.ObjectId:
            return cls._validate(v)

        @classmethod
        def __modify_schema__(cls, field_schema: dict[str, Any]) -> None:
            field_schema.update(type="string")

    @classmethod
    def _validate(cls, v: bson.ObjectId | bytes) -> bson.ObjectId:
        if not bson.ObjectId.is_valid(v):
            raise ValueError("Invalid object id")
        return bson.ObjectId(v)



from typing import TypeVar

from fastapi import Query
from fastapi._compat import field_annotation_is_scalar
from pydantic import BaseModel, Field, create_model

from compose import compat

Q = TypeVar("Q", bound=BaseModel)


def to_query(q: type[Q], /) -> type[Q]:
    field_args = (
        "title",
        "alias",
        "default",
        "default_factory",
        "description",
    )

    if compat.IS_PYDANTIC_V2:
        field_args = (*field_args, "annotation")
        field_definitions = {
            field_name: (
                field_info.annotation,
                (
                    field_info
                    if field_annotation_is_scalar(field_info.annotation)
                    else Field(Query(**{arg: getattr(field_info, arg, None) for arg in field_args}))
                ),
            )
            for field_name, field_info in q.model_fields.items()
        }
        return create_model(f"{q.__name__}Query", **field_definitions, __base__=q)

    else:
        field_definitions = {
            field_name: (
                field.outer_type_,
                (
                    field.field_info
                    if field_annotation_is_scalar(field.outer_type_)
                    else Field(Query(**{arg: getattr(field, arg, None) for arg in field_args}))
                ),
            )
            for field_name, field in q.__fields__.items()
        }

        class Child(q):
            class Config:
                arbitrary_types_allowed = True

        return create_model(
            f"{q.__name__}Query",
            **field_definitions,
            __base__=Child,
        )



import enum
import os
from collections.abc import Callable
from typing import Any, Self


def enum_values(e: type[enum.Enum], /) -> list[Any]:
    return [member.value for _, member in e.__members__.items()]


def create_env_getter(key: str, /) -> Callable[[], str | None]:
    def getter() -> str | None:
        return os.getenv(key)

    return getter


default_env_getter = create_env_getter("APP_ENV")


class AppEnv(enum.StrEnum):
    TEST = enum.auto()
    LOCAL = enum.auto()
    DEV = enum.auto()
    STG = enum.auto()
    PRD = enum.auto()

    @classmethod
    def current(cls, env_getter: Callable[[], str | None] = default_env_getter) -> Self:
        if (env := env_getter()) is None or env not in set(enum_values(cls)):
            raise ValueError(f"Invalid value for {cls.__name__}: {env}")

        return cls(env)



from collections.abc import Callable
from pathlib import Path

import pytest


def default_marker_getter(item: pytest.Function) -> pytest.MarkDecorator:
    test_type_to_marker = {
        marker.name: marker
        for marker in [pytest.mark.unit, pytest.mark.integration, pytest.mark.e2e]
    }
    test_types = set(test_type_to_marker.keys())
    default_test_type = pytest.mark.unit.name

    node_path = item.nodeid.split("::")[0]
    parts = Path(node_path).parts
    return next((part for part in parts if part in test_types), default_test_type)


def add_test_type_markers(
    items: list[pytest.Function],
    marker_getter: Callable[[pytest.Function], pytest.MarkDecorator] = default_marker_getter,
) -> None:
    """

    Examples:
        >>> from compose import testing
        >>> import pytest
        >>> def pytest_collection_modifyitems(items: list[pytest.Function]) -> None:
        >>>     testing.add_test_type_markers(items)

    """

    for item in items:
        item.add_marker(marker_getter(item))



from collections.abc import Callable

from . import utils
from .base import LogicalOperator
from .types import ListExpression


def _expression_factory(
    mongo_operator: str,
) -> Callable[[LogicalOperator], dict[str, ListExpression]]:
    def expression(self: LogicalOperator) -> dict[str, ListExpression]:
        expressions = [expr for op in self.ops if (expr := op.expression())]
        return {mongo_operator: expressions} if expressions else {}

    return expression


def create_logical_operator(name: str, mongo_operator: str) -> type[LogicalOperator]:
    return utils.create_operator(
        name=name, base=(LogicalOperator,), expression_factory=_expression_factory(mongo_operator)
    )


And = create_logical_operator(name="And", mongo_operator="$and")
Or = create_logical_operator(name="Or", mongo_operator="$or")
Nor = create_logical_operator(name="Nor", mongo_operator="$nor")



import enum
from collections.abc import Callable

import pytest

from . import param


def create_enum_flag_property_test(
    e: type[enum.Enum],
    f: Callable[[enum.Enum], bool],
    /,
    truthy: list[enum.Enum] | tuple[enum.Enum] | set[enum.Enum],
) -> Callable[[enum.Enum, bool], None]:
    """

    Examples:
        ```python
        import enum

        class Status(enum.StrEnum):
            SUCCEEDED = enum.auto()
            FAILED = enum.auto()

            @property
            def is_succeeded(self) -> bool:
                return self == Status.SUCCEEDED

        test_status_is_succeeded = compose.testing.create_enum_flag_property_test(
            Status,
            Status.is_succeeded.fget,
            truthy={Status.SUCCEEDED},
        )
        ```

    """

    def _test(value: enum.Enum, expected: bool) -> None:
        assert f(value) is expected

    return pytest.mark.parametrize(
        "value, expected", param.parametrize_enum_flag_property_test(e=e, truthy=truthy)
    )(_test)



from . import func
from .base import Evaluable, ListExpression, Operator


class Pipeline(Operator):
    def __init__(self, *ops: Operator):
        self.ops = list(ops)

    def expression(self) -> ListExpression:
        return Evaluable(func.Flatten(func.Filter(self.ops, func.NonEmpty()))).expression()



from collections.abc import Callable, Generator
from typing import Any, Self, TypeVar

from compose import compat

from . import CoreSchemaGettable

T = TypeVar("T")


def caster(factory: Callable[[Any], T], /) -> Callable[[Any], T]:
    def _cast(v: Any) -> T:
        return factory(v)

    return _cast


class Str(str, CoreSchemaGettable[str]):
    @classmethod
    def __get_validators__(cls) -> Generator[[Callable[[Any], Self]], None, None]:
        yield compat.str_validator
        yield caster(cls)


class Int(int, CoreSchemaGettable[int]):
    @classmethod
    def __get_validators__(cls) -> Generator[[Callable[[Any], Self]], None, None]:
        yield compat.int_validator
        yield caster(cls)


class Float(float, CoreSchemaGettable[float]):
    @classmethod
    def __get_validators__(cls) -> Generator[[Callable[[Any], Self]], None, None]:
        yield compat.float_validator
        yield caster(cls)


class StrList(list[str], CoreSchemaGettable[list[str]]):
    @classmethod
    def __get_validators__(cls) -> Generator[[Callable[[Any], Self]], None, None]:
        yield compat.list_validator
        yield caster(cls)


class IntList(list[int], CoreSchemaGettable[list[int]]):
    @classmethod
    def __get_validators__(cls) -> Generator[[Callable[[Any], Self]], None, None]:
        yield compat.list_validator
        yield caster(cls)



from . import _signal as signal
from . import (
    command,
    compat,
    concurrent,
    dependency,
    entity,
    enums,
    event,
    exceptions,
    field,
    messaging,
    pagination,
    query,
    repository,
    result,
    schema,
    settings,
    stream,
    types,
    uow,
)
from .container import BaseModel, TimeStampedModel

__all__ = [
    "BaseModel",
    "TimeStampedModel",
    "entity",
    "field",
    "schema",
    "repository",
    "query",
    "types",
    "command",
    "event",
    "dependency",
    "compat",
    "pagination",
    "result",
    "uow",
    "messaging",
    "concurrent",
    "signal",
    "exceptions",
    "fastapi",
    "settings",
    "enums",
    "stream",
]

try:
    from . import logging  # noqa: F401

    __all__.append("logging")
except ImportError:
    pass

try:
    from . import testing  # noqa: F401

    __all__.append("testing")
except ImportError:
    pass

try:
    from .framework import fastapi  # noqa: F401

    __all__.append("fastapi")
except ImportError:
    pass



from .consumer import MessageConsumer
from .consumer_runner import ThreadMessageConsumerRunner
from .messagebus import MessageBus
from .model import EventMessage, SqsEventMessage
from .publisher import EventPublisher
from .queue import MessageQueue
from .signal_handler import DefaultSignalHandler, SignalHandler, ThreadSignalHandler

__all__ = [
    "EventMessage",
    "SqsEventMessage",
    "MessageQueue",
    "MessageConsumer",
    "ThreadMessageConsumerRunner",
    "MessageBus",
    "EventPublisher",
    "SignalHandler",
    "DefaultSignalHandler",
    "ThreadSignalHandler",
]

try:
    from .queue import SqsMessageQueue  # noqa: F401

    __all__.append("SqsMessageQueue")
except ImportError:
    pass



import asyncio
import json
import logging
from collections.abc import Callable
from typing import Literal, TypeAlias

from . import model
from .messagebus import MessageBus
from .queue import MessageQueue
from .signal_handler import DefaultSignalHandler, SignalHandler

HookEventType = Literal[
    "on_start",
    "on_receive",
    "on_receive_error",
    "on_consume",
    "on_consume_error",
    "on_shutdown",
]
HookArgType: TypeAlias = str | model.EventMessage | Exception
Hook: TypeAlias = Callable[[HookArgType], None]

logger = logging.getLogger("compose")


def log_event_message(log_message: str, message: model.EventMessage) -> None:
    logger.info(
        f"{log_message}: {json.dumps(message.encode(), ensure_ascii=False)}",
        extra={"event_message": message.encode()},
    )


def log_exception(log_message: str, exc: Exception) -> None:
    logger.exception(log_message, exc_info=exc, stack_info=True)


DEFAULT_HOOKS = {
    "on_start": [logger.info],
    "on_receive": [lambda message: log_event_message("Received message", message)],
    "on_receive_error": [lambda exc: log_exception("Failed to receive message", exc)],
    "on_consume": [lambda message: log_event_message("Consumed message", message)],
    "on_consume_error": [lambda exc: log_exception(f"Failed to consume message due to {exc}", exc)],
    "on_shutdown": [logger.info],
}


class MessageConsumer:
    def __init__(
        self,
        messagebus: MessageBus,
        message_queue: MessageQueue,
        hooks: dict[HookEventType, list[Hook]] | None = None,
        signal_handler: SignalHandler | None = None,
    ):
        self.messagebus = messagebus
        self.message_queue = message_queue
        self.hooks = DEFAULT_HOOKS | (hooks or {})
        self.signal_handler = signal_handler or DefaultSignalHandler()

        self._default_hook = lambda _: None

    async def run(self) -> None:
        self._execute_hook("on_start", "MessageConsumer started")

        while not self.signal_handler.received_signal:
            try:
                message = self.message_queue.peek()
            except Exception as exc:
                self._execute_hook("on_receive_error", exc)
                continue

            if message is None:
                continue
            self._execute_hook("on_receive", message)

            try:
                await asyncio.create_task(self.consume(message))
            except Exception as exc:
                self._execute_hook("on_consume_error", exc)
                continue
            self._execute_hook("on_consume", message)

    async def consume(self, message: model.EventMessage) -> None:
        await self.messagebus.handle_event(message.body)
        self.message_queue.delete(message)

    def _execute_hook(self, hook_event_type: HookEventType, arg: HookArgType, /) -> None:
        for hook in self.hooks.get(hook_event_type, [self._default_hook]):
            hook(arg)

    def shutdown(self) -> None:
        self._execute_hook("on_shutdown", "MessageConsumer shutting down")
        self.signal_handler.handle()



from typing import Any

from .. import utils
from ..base import ComparisonOperator
from ..types import DictExpression


class AEqual(ComparisonOperator):
    def __init__(self, field: str, value: Any | None = None):
        super().__init__(field=field, value=value)

    def expression(self) -> DictExpression:
        return {"$eq": [self.field, self.value]}


AEq = utils.create_general_aggregation_operator(name="AEq", mongo_operator="$eq")
ANe = utils.create_general_aggregation_operator(name="ANe", mongo_operator="$ne")
AGt = utils.create_general_aggregation_operator(name="AGt", mongo_operator="$gt")
AGte = utils.create_general_aggregation_operator(name="AGte", mongo_operator="$gte")
ALt = utils.create_general_aggregation_operator(name="ALt", mongo_operator="$lt")
ALte = utils.create_general_aggregation_operator(name="ALte", mongo_operator="$lte")



from typing import Any

import pytest


def get_fixture_param(request: pytest.FixtureRequest, default: Any | None = None) -> Any:
    return getattr(request, "param", default)


def get_fixture_value_from_param(request: pytest.FixtureRequest) -> Any:
    return request.getfixturevalue(get_fixture_param(request))



import asyncio
import collections
import functools
import re
from collections.abc import Callable
from typing import Protocol, Self

from compose.dependency.wiring import create_lazy_resolver
from compose.event import Event


class EventHandler(Protocol):
    def handle(self, evt: Event) -> None:
        ...


class MessageBus:
    def __init__(self, dependency_resolver: Callable[[str], EventHandler]):
        self.dependency_resolver = dependency_resolver

        self._event_handlers: dict[str, set[str]] = collections.defaultdict(set)

    @classmethod
    def with_container(cls, container_path: str) -> Self:
        container_path_pattern = re.compile("^(?P<module_path>.+):(?P<container_name>.+)$")
        if container_path_pattern.match(container_path) is None:
            raise ValueError(
                f"Invalid container path: {container_path}. "
                "Must be in the format `module.path:ContainerName`"
            )

        return cls(dependency_resolver=create_lazy_resolver(container_path))

    async def handle_event(self, evt: Event) -> None:
        handler_names = self._event_handlers.get(evt.__class__.__name__, set())
        event_loop = asyncio.get_running_loop()
        for handler_name in handler_names:
            handler = self.dependency_resolver(handler_name)
            await event_loop.run_in_executor(None, functools.partial(handler.handle, evt))

    def register(self, event_cls: type[Event]) -> Callable[[EventHandler], EventHandler]:
        def wrapper(handler_cls: EventHandler) -> EventHandler:
            event_name = event_cls.__name__
            handler_name = handler_cls.__name__

            registered_handler_names = self._event_handlers.get(event_cls.__name__, set())
            if handler_cls.__name__ in registered_handler_names:
                raise ValueError(
                    f"Handler {handler_name} already registered for event {event_name}"
                )

            self._event_handlers[event_name].add(handler_name)
            return handler_cls

        return wrapper



import http
import urllib.parse
from collections.abc import Mapping
from typing import Any, Self

from fastapi import Response
from fastapi.responses import StreamingResponse
from starlette.background import BackgroundTask
from starlette.responses import ContentStream


class NoContentResponse(Response):
    def __init__(
        self,
        content: Any = None,
        headers: Mapping[str, str] | None = None,
        media_type: str | None = None,
        background: BackgroundTask | None = None,
    ) -> None:
        super().__init__(
            content=content,
            headers=headers,
            media_type=media_type,
            status_code=http.HTTPStatus.NO_CONTENT,
            background=background,
        )


class ZipStreamingResponse(StreamingResponse):
    default_media_type = "application/zip"

    @classmethod
    def with_filename(
        cls,
        content: ContentStream,
        filename: str,
        background: BackgroundTask | None = None,
    ) -> Self:
        return cls(
            content=content,
            media_type=cls.default_media_type,
            headers={"Content-Disposition": f"attachment; filename={urllib.parse.quote(filename)}"},
            background=background,
        )



from typing import Any, Callable, Type

from pydantic import BaseModel as PydanticBaseModel

from compose import compat

SchemaExtraCallable = Callable[[dict[str, Any], Type[PydanticBaseModel]], None]


if compat.IS_PYDANTIC_V2:

    def schema_excludes(*excludes: str) -> SchemaExtraCallable:
        """OpenAPI 문서에서 특정 필드를 제외합니다. 외부에 노출하지 않아도 되는 필드를 문서에서 제외할 때 사용합니다."""

        def wrapper(schema: dict[str, Any], model_class: Type[PydanticBaseModel]) -> None:
            for exclude in excludes:
                schema.get("properties", {}).pop(exclude)

        return wrapper

    def schema_by_field_name() -> SchemaExtraCallable:
        """OpenAPI 문서에서 스키마의 필드명을 `field.alias`가 아닌 원래의 필드명으로 표시합니다.
        FastAPI는 `field.alias`를 기본 필드명으로 사용하므로 스키마를 순회하며 스키마 필드명을 원 필드명으로 수정합니다.

        References:
            https://github.com/tiangolo/fastapi/issues/1810#issuecomment-895126406
        """

        def wrapper(schema: dict[str, Any], model_class: Type[PydanticBaseModel]) -> None:
            updated = dict()
            for field_name, field in model_class.model_fields.items():
                alias = field.alias or field_name
                prop = schema.get("properties", {}).get(alias)
                updated[field_name] = prop

            schema |= dict(properties=updated)

        return wrapper

else:

    def schema_excludes(*excludes: str) -> SchemaExtraCallable:
        """OpenAPI 문서에서 특정 필드를 제외합니다. 외부에 노출하지 않아도 되는 필드를 문서에서 제외할 때 사용합니다."""

        def wrapper(schema: dict[str, Any], model_class: type[PydanticBaseModel]) -> None:
            for exclude in excludes:
                schema.get("properties", {}).pop(exclude)

        return wrapper

    def schema_by_field_name() -> SchemaExtraCallable:
        """OpenAPI 문서에서 스키마의 필드명을 `field.alias`가 아닌 원래의 필드명으로 표시합니다.
        FastAPI는 `field.alias`를 기본 필드명으로 사용하므로 스키마를 순회하며 스키마 필드명을 원 필드명으로 수정합니다.

        References:
            https://github.com/tiangolo/fastapi/issues/1810#issuecomment-895126406
        """

        def wrapper(schema: dict[str, Any], model_class: type[PydanticBaseModel]) -> None:
            updated = dict()
            for field in model_class.__fields__.values():
                field_name = field.name
                alias = field.alias or field_name
                prop = schema.get("properties", {}).get(alias)
                updated[field_name] = prop

            schema.update({"properties": updated})

        return wrapper



from __future__ import annotations

import abc
import functools
import operator
from collections.abc import Callable
from typing import Any, ClassVar, Generic, Self, TypeVar

from .types import DictExpression, ListExpression

T = TypeVar("T")


class Operator:
    @abc.abstractmethod
    def expression(self) -> Any:
        raise NotImplementedError

    def fmap(self, op: Callable[..., Operator]) -> Operator:
        return op(self.expression())


class ComparisonOperator(Operator):
    def __init__(self, field: str, value: Any | None = None):
        self.field = field
        self.value = value

    @abc.abstractmethod
    def expression(self) -> dict[str, Any]:
        raise NotImplementedError


class LogicalOperator(Operator):
    def __init__(self, *ops: Operator):
        self.ops = list(ops)

    @abc.abstractmethod
    def expression(self) -> dict[str, ListExpression]:
        raise NotImplementedError


class GeneralAggregationOperator(Operator):
    mongo_operator: ClassVar[str] = ""

    def __init__(self, *expressions: Any):
        self.expressions = list(expressions)

    def expression(self) -> DictExpression:
        return Evaluable({self.mongo_operator: self.expressions}).expression()


class Stage(Operator, Generic[T]):
    @abc.abstractmethod
    def expression(self) -> T:
        raise NotImplementedError


class Merge(Operator):
    def __init__(self, *ops: Operator, initial: Any):
        self.ops = list(ops)
        self.initial = initial

    def expression(self) -> Any:
        return functools.reduce(operator.or_, [op.expression() for op in self.ops], self.initial)

    @classmethod
    def dict(cls, *ops: Operator) -> Self:
        return cls(*ops, initial={})


class Evaluable(Operator):
    def __init__(self, op: Any):
        self.op = op

    def expression(self) -> Any:
        match self.op:
            case Operator():
                return self.op.expression()
            case dict():
                return {k: Evaluable(v).expression() for k, v in self.op.items()}
            case list():
                return [Evaluable(v).expression() for v in self.op]
            case _:
                return self.op



try:
    import fastapi  # noqa: F401
except ImportError:
    raise ImportError("Install `fastapi` to use fastapi helpers")

from .endpoint import health_check
from .exception_handler import (
    ExceptionHandler,
    ExceptionHandlerInfo,
    create_exception_handler,
)
from .openapi import openapi_tags
from .param import to_query
from .response import NoContentResponse, ZipStreamingResponse
from .security import APIKeyAuth, HTTPBasicAuth

__all__ = [
    "ExceptionHandler",
    "ExceptionHandlerInfo",
    "create_exception_handler",
    "to_query",
    "health_check",
    "HTTPBasicAuth",
    "APIKeyAuth",
    "openapi_tags",
    "NoContentResponse",
    "ZipStreamingResponse",
]



from .model import (
    model_dump,
    model_dump_json,
    model_schema,
    model_validate,
    model_validate_json,
    validate_obj,
)
from .utils import IS_PYDANTIC_V2

if IS_PYDANTIC_V2:
    from pydantic.v1.validators import *  # noqa: F401, F403
else:
    from pydantic.validators import *  # noqa: F401, F403


__all__ = [
    "IS_PYDANTIC_V2",
    "validate_obj",
    "model_schema",
    "model_validate",
    "model_validate_json",
    "model_dump",
    "model_dump_json",
]



import functools
from collections.abc import Callable, Generator
from typing import Any, TypeVar, get_type_hints

T = TypeVar("T")
RT = TypeVar("RT")


def descendants_of(cls: type[T]) -> Generator[type[T], None, None]:
    stack = cls.__subclasses__()
    while stack:
        current_cls = stack.pop()
        yield current_cls
        stack.extend(current_cls.__subclasses__())


def unordered_partial(p: functools.partial[RT], t: T) -> Callable[..., RT]:
    type_hints = get_type_hints(p.func)

    exclude_keys = {*p.keywords.keys(), "return"}
    candidates = [k for k, v in type_hints.items() if v == t and k not in exclude_keys]

    if not candidates or len(candidates) > 1:
        raise TypeError(
            f"Cannot inject argument of type {t} into {p}. "
            f"Expected exactly one argument of type {t}, "
            f"but found {len(candidates)}: {candidates}"
        )

    arg_name = candidates[0]

    def wrapper(arg: Any) -> T:
        return p(**{arg_name: arg})

    return functools.partial(wrapper)


def ident(x: T) -> T:
    return x



import abc

from .. import model


class MessageQueue(abc.ABC):
    @abc.abstractmethod
    def push(self, message: model.EventMessage) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def peek(self) -> model.EventMessage | None:
        raise NotImplementedError

    @abc.abstractmethod
    def delete(self, message: model.EventMessage) -> None:
        raise NotImplementedError



import enum
import functools
import importlib
import inspect
from collections.abc import Callable, Iterable
from typing import Any, NotRequired, Protocol, TypedDict, TypeVar, Unpack

from dependency_injector import containers, providers
from dependency_injector.wiring import Provide

T = TypeVar("T")
Container = type[containers.Container] | containers.Container

DEFAULT_RESOLVABLE_PROVIDER_TYPES = (providers.Factory, providers.Singleton)


class Wirer(Protocol):
    def __call__(
        self,
        container: containers.Container,
        modules: Iterable[str] | None = None,
        from_package: str | None = None,
    ) -> None:
        ...


def create_wirer(packages: Iterable[str]) -> Wirer:
    def wire_container(
        container: containers.Container,
        modules: Iterable[str] | None = None,
        from_package: str | None = None,
    ) -> None:
        container.check_dependencies()
        container.wire(modules=modules, packages=packages, from_package=from_package)

    return wire_container


def resolve_by_name_from_container_provider(
    name: str,
    container: providers.Container,
    provider_types: Iterable[type[providers.Provider]],
) -> providers.Factory:
    if not isinstance(name, str):
        raise ValueError("`name` must be string")

    for provider_name, provider in container.providers.items():
        if not isinstance(provider, (providers.Container, *provider_types)):
            continue

        if provider_name == name:
            return provider

        if isinstance(provider, providers.Container):
            return resolve_by_name_from_container_provider(
                name=name,
                container=provider,
                provider_types=provider_types,
            )


def resolve_by_name(
    name: str,
    container: Container,
    provider_types: Iterable[type[providers.Provider]],
) -> providers.Factory:
    if not isinstance(name, str):
        raise ValueError("`name` must be string")

    for provider_name, provider in container.providers.items():
        if not isinstance(provider, (providers.Container, *provider_types)):
            continue

        if provider_name == name:
            return provider

        if isinstance(provider, providers.Container):
            result = resolve_by_name_from_container_provider(
                name=name,
                container=provider,
                provider_types=provider_types,
            )
            if result is not None:
                return result

    raise ValueError(f"Cannot find {name} from given container")


def resolve_by_object_name(
    name: str,
    container: Container,
    provider_types: Iterable[type[providers.Provider]],
) -> Any:
    candidates: list[providers.Factory] = []
    for provider in container.traverse([*provider_types]):
        if not inspect.isclass(provider.cls):
            continue

        if provider.cls.__name__.split(".")[-1] == name:
            candidates.append(provider)

    if not candidates:
        raise ValueError(f"Cannot find {name} from given container")

    if len(candidates) > 1:
        raise ValueError(f"Cannot resolve {name} since there are multiple candidates")

    return candidates[0]()


class ConflictResolution(str, enum.Enum):
    FIRST = "first"
    ERROR = "error"


@functools.lru_cache(32)
def resolve(
    type_: type[Any] | str,
    container: Container,
    provider_types: Iterable[type[providers.Provider]] = DEFAULT_RESOLVABLE_PROVIDER_TYPES,
    *,
    name: str | None = None,
    conflict_resolution: ConflictResolution = ConflictResolution.FIRST,
) -> providers.Factory:
    """
    의존성 전체 등록 경로를 참조하지 않고 의존성을 해결합니다. 다른 패키지의 의존성을 참조하는 경우
    의존 대상 선언 경로에 깊게 의존하는 것을 방지합니다. `container_cls`는 최상위 컨테이너일수도,
    의존성이 등록된 (하위) 컨테이너일수도 있습니다. 클래스 대상으로만 작동합니다.
    """
    if not (inspect.isclass(type_) or isinstance(type_, str)):
        raise ValueError("Only class or string can be resolved")

    if isinstance(type_, str):
        return resolve_by_name(name=type_, container=container, provider_types=provider_types)

    candidates = []
    for provider in container.traverse([*provider_types]):
        provider_cls = provider.cls
        if not (inspect.isclass(provider_cls) or inspect.ismethod(provider_cls)):
            continue

        cls = provider_cls.__self__ if inspect.ismethod(provider_cls) else provider_cls
        if cls.__name__ == type_.__name__:
            candidates.append(provider)

    if not candidates:
        raise ValueError(f"Cannot find {type_.__name__} from given container")

    if len(candidates) > 1 and name is None and conflict_resolution == ConflictResolution.ERROR:
        type_name = type_.__name__ if inspect.isclass(type_) else type_
        raise ValueError(
            f"Cannot resolve {type_name} since there are multiple candidates. "
            f"You must specify `name` argument to resolve dependency"
        )

    if len(candidates) == 1:
        return candidates[0]

    if name is None and conflict_resolution == ConflictResolution.FIRST:
        return candidates[0]

    return resolve_by_name(name=name, container=container, provider_types=provider_types)


def provide(
    type_: type[T],
    from_: type[containers.Container],
    /,
    *,
    provider_types: Iterable[type[providers.Provider]] = DEFAULT_RESOLVABLE_PROVIDER_TYPES,
    name: str | None = None,
    conflict_resolution: ConflictResolution = ConflictResolution.FIRST,
) -> Provide[T]:
    return Provide[
        resolve(
            type_=type_,
            container=from_,
            provider_types=provider_types,
            name=name,
            conflict_resolution=conflict_resolution,
        )
    ]


def create_resolver(container: Container) -> Callable[[str], Any]:
    def resolver(name: str) -> Any:
        return resolve(type_=name, container=container)

    return resolver


def create_lazy_resolver(container_path: str) -> Callable[[str], Any]:
    def resolver(object_name: str) -> Any:
        module_path, container_name = container_path.split(":")
        try:
            container = importlib.import_module(module_path)
        except ImportError:
            raise ImportError(f"Cannot not import module {module_path}")

        if (container_cls := getattr(container, container_name, None)) is None:
            raise ValueError(f"Cannot find container {container_name} in {module_path}")

        return resolve_by_object_name(
            name=object_name,
            container=container_cls,
            provider_types=DEFAULT_RESOLVABLE_PROVIDER_TYPES,
        )

    return resolver


class ProvideExtraArgs(TypedDict):
    name: NotRequired[str]
    conflict_resolution: NotRequired[ConflictResolution]


def create_provider(
    container: type[containers.Container],
    provider_types: Iterable[type[providers.Provider]] = DEFAULT_RESOLVABLE_PROVIDER_TYPES,
) -> Callable[[type[T], Unpack[ProvideExtraArgs]], Provide[T]]:
    def provider(
        type_: type[T],
        /,
        name: str | None = None,
        conflict_resolution: ConflictResolution = ConflictResolution.FIRST,
    ) -> Provide[T]:
        return provide(
            type_,
            container,
            provider_types=provider_types,
            name=name,
            conflict_resolution=conflict_resolution,
        )

    return provider



from typing import Any, Self

from ..base import Evaluable, Operator
from ..types import DictExpression, MongoKeyword, _String


class Map(Operator):
    def __init__(self, input_: Any, as_: str, in_: Any):
        self.input = input_
        self.as_ = _String(as_)
        self.in_ = in_

    def expression(self) -> DictExpression:
        return {
            "$map": {
                MongoKeyword.from_py(field): Evaluable(value).expression()
                for field, value in self.__dict__.items()
            }
        }


class Size(Operator):
    def __init__(self, expression: Any):
        self._expression = expression

    def expression(self) -> DictExpression:
        return {"$size": Evaluable(self._expression).expression()}


class Filter(Operator):
    def __init__(self, input_: Any, as_: str, cond: Any, limit: Any | None = None):
        self.input = input_
        self.as_ = _String(as_)
        self.cond = cond
        self.limit = limit

    def expression(self) -> DictExpression:
        return {
            "$filter": {
                MongoKeyword.from_py(field): Evaluable(value).expression()
                for field, value in self.__dict__.items()
                if value is not None
            },
        }


class Reduce(Operator):
    def __init__(self, input_: Any, initial_value: Any, in_: Any):
        self.input = input_
        self.initial_value = initial_value
        self.in_ = in_

    def expression(self) -> DictExpression:
        return {
            "$reduce": {
                MongoKeyword.from_py(field): Evaluable(value).expression()
                for field, value in self.__dict__.items()
            }
        }

    @classmethod
    def list(cls, input_: Any, in_: Any) -> Self:
        return cls(input_=input_, initial_value=[], in_=in_)

    @classmethod
    def int(cls, input_: Any, in_: Any) -> Self:
        return cls(input_=input_, initial_value=0, in_=in_)



from collections.abc import Generator
from typing import IO, TypeVar

T = TypeVar("T")


def chunk_reader(f: IO[T], chunk_size: int = 1024 * 1024) -> Generator[T, None, None]:
    while chunk := f.read(chunk_size):
        yield chunk



from typing import Any

from ..base import Evaluable, GeneralAggregationOperator, Operator
from ..types import DictExpression


class Concat(GeneralAggregationOperator):
    mongo_operator = "$concat"


class Split(Operator):
    def __init__(self, expr: Any, delimiter: str):
        self.expr = expr
        self.delimiter = delimiter

    def expression(self) -> DictExpression:
        return Evaluable({"$split": [self.expr, self.delimiter]}).expression()



from typing import Any

from ..base import Evaluable, Operator
from ..types import DictExpression


class ToString(Operator):
    def __init__(self, expr: Any):
        self.expr = expr

    def expression(self) -> DictExpression:
        return Evaluable({"$toString": self.expr}).expression()


class ToBool(Operator):
    def __init__(self, expression: Any):
        self._expression = expression

    def expression(self) -> DictExpression:
        return Evaluable({"$toBool": self._expression}).expression()


class ToInt(Operator):
    def __init__(self, expression: Any):
        self._expression = expression

    def expression(self) -> DictExpression:
        return Evaluable({"$toInt": self._expression}).expression()



import logging

from ..event import Event
from .model import EventMessage
from .queue import MessageQueue

logger = logging.getLogger("compose")


class EventPublisher:
    def __init__(self, message_queue: MessageQueue):
        self.message_queue = message_queue

    def publish(self, evt: Event) -> None:
        self.message_queue.push(EventMessage(body=evt))

        logger.info(f"Published event: {evt.__class__.__name__}")
        logger.debug(f"Event: {evt.encode()}")



from typing import Callable, Generic, TypeAlias, TypeVar, final

T = TypeVar("T")
E = TypeVar("E", bound=Exception)


class _Result(Generic[T, E]):
    def unwrap(self) -> T:
        if isinstance(self, Ok):
            return self.value

        raise ValueError("Cannot unwrap value from `Err`")

    def unwrap_err(self) -> E:
        if isinstance(self, Err):
            return self.error

        raise ValueError("Cannot unwrap error from `Ok`")

    def unwrap_or(self, default: T) -> T:
        if isinstance(self, Ok):
            return self.value

        return default


@final
class Ok(_Result[T, E]):
    __match_args__ = ("value",)

    def __init__(self, value: T):
        self.value = value


@final
class Err(_Result[T, E]):
    __match_args__ = ("error",)

    def __init__(self, error: E):
        self.error = error


Result: TypeAlias = Ok | Err


def to_result(func: Callable[..., T]) -> Result:
    try:
        return Ok(func())
    except Exception as exc:
        return Err(exc)



from typing import Any, ClassVar, TypeVar

from pydantic import BaseModel

from . import compat, container, field, types

ModelType = TypeVar("ModelType", bound=BaseModel)


class Entity(container.TimeStampedModel):
    id: types.PyObjectId = field.IdField(default_factory=types.PyObjectId)

    updatable_fields: ClassVar[set[str]] = set()

    if compat.IS_PYDANTIC_V2:

        @classmethod
        def __pydantic_init_subclass__(cls, **kwargs: Any) -> None:
            super().__pydantic_init_subclass__(**kwargs)

            fields = set(cls.model_fields.keys())
            if diff := set(cls.updatable_fields) - fields:
                raise ValueError(f"`updatable_fields` must be subset of {fields}, but got {diff}")

    else:

        def __init_subclass__(cls, **kwargs: Any) -> None:
            super().__init_subclass__(**kwargs)

            fields = set(cls.__fields__.keys())
            if diff := set(cls.updatable_fields) - fields:
                raise ValueError(f"`updatable_fields` must be subset of {fields}, but got {diff}")

            id_field = cls.__fields__.pop("id")
            cls.__fields__ = dict(id=id_field) | cls.__fields__

    def update(self, **kwargs: Any) -> None:
        for key, value in kwargs.items():
            if key not in self.updatable_fields:
                continue

            setattr(self, key, value)



from typing import Any

from .. import utils
from ..base import Evaluable, Operator

AIn = utils.create_general_aggregation_operator(name="AIn", mongo_operator="$in")


class ArrayElemAt(Operator):
    def __init__(self, array: Any, index: int, /):
        self.array = array
        self.index = index

    def expression(self) -> Any:
        return {"$arrayElemAt": [Evaluable(self.array).expression(), self.index]}



import abc
from typing import Any

from pydantic import Field

from .. import base


class MongoQuery(base.Query, abc.ABC):
    @abc.abstractmethod
    def to_query(self) -> list[dict[str, Any]]:
        raise NotImplementedError


class MongoFilterQuery(MongoQuery):
    page: int | None = Field(None, ge=1)
    per_page: int | None = Field(None, ge=1)

    @abc.abstractmethod
    def to_query(self) -> list[dict[str, Any]]:
        raise NotImplementedError

    @property
    def can_paginate(self) -> bool:
        return self.page is not None and self.per_page is not None


MongoOffsetFilterQuery = MongoFilterQuery



import concurrent.futures
import functools
from collections.abc import Callable
from typing import TypeVar

K = TypeVar("K")
T = TypeVar("T")


def execute_in_pool(
    pool_factory: Callable[[], concurrent.futures.Executor],
    funcs: dict[K, functools.partial[T]],
    timeout: int | None = None,
) -> dict[K, T]:
    result = {}
    with pool_factory() as executor:
        future_to_key = dict()
        for key, func in funcs.items():
            future = executor.submit(func)
            future_to_key[future] = key

        for future in concurrent.futures.as_completed(future_to_key, timeout=timeout):
            result[future_to_key[future]] = future.result()

    return result



from typing import Any

from .base import Evaluable, Operator
from .types import DictExpression


class Expr(Operator):
    def __init__(self, op: Any):
        self.op = op

    def expression(self) -> DictExpression:
        return Evaluable({"$expr": self.op}).expression()
